/*
 * A modified Floyd-Steinberg dithering that produces tuti-interweavable bitmaps
 * 2017-01-31: coded by Abdalla G. M. Ahmed
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string>
#include <getopt.h>

const char *USAGE_MESSAGE = "Usage: %s <pgm-file>\n"
    "The pgm file must not contain comments\n"
    "Options\n:"
    "-n     produce regular (not Tuti) halftone.\n"
;

int main(int argc, char **argv) {
    int opt;                                                                    // For use by getopt, the command line options utility
    bool tutiHalftone = true;                                                   // Generate Tuti halftone by default
    while ((opt = getopt(argc, argv, "n")) != -1) {                             // Modify default settings with command line options
        switch (opt) {
            case 'n': tutiHalftone = false; break;
            default: fprintf(stderr, USAGE_MESSAGE, argv[0]); exit(1);
        }
    }
    if (optind > argc - 1) {
        fprintf(stderr, USAGE_MESSAGE, argv[0]); exit(1);
    }
    std::string fileName = argv[optind];
    FILE *pgmfile = fopen(fileName.c_str(), "r");
    unsigned int w, h, grades;
    // --------------- Read input image --------------------
    fscanf(pgmfile, "P2 %d %d %d ", &w, &h, &grades);
    int zero(0);                                                                // The lowest color level in the output
    if (tutiHalftone) {                                                         // We need to modify the tone range to (25~75%)
        zero = grades / 2;
        grades *= 2;
    }
    int *bmp = new int[w*h];
    for (int i = 0; i < w*h; i++) {
        fscanf(pgmfile, " %d", &bmp[i]);
        if (feof(pgmfile)) {
            fprintf(stderr, "Failed to load image\n");
            exit(1);
        }
        bmp[i] += zero;
    }
    fclose(pgmfile);
    // --------- Compute and write output image -----------
    fileName.replace(fileName.length() - 3, 3, "pbm");
    FILE *pbmfile = fopen(fileName.c_str(), "w");
    fprintf(pbmfile, "P1\n%d %d\n", w, h);                                      // Print output file header
    char *row = new char[w];                                                    // To hold a single row of output; needed because of serpentine scanning.
    int threshold = grades / 2, error;
    for (int y = 0; y < h; y++) {                                               // -------------- Dither and print output raster -----------------
        if ((y&1) == 0) {                                                       // Serpentine scan
            for (int x = 0; x < w; x++) {
                int i = y*w + x;
                if (bmp[i] < threshold) {
                    row[x] = '1';                                               // Output black pixel
                    error = bmp[i];                                             // Positive error: carry-on to neighbors
                } else {
                    row[x] = '0';                                               // Output white pixel
                    error = bmp[i] - grades;                                    // Negative error: to be borrowed from neighbors
                }
                if (tutiHalftone) {
                    if (((x&1) == 0) && (row[x] != '1')) {                          // Even-even pixel must be black; if it's not fix it
                        row[x] = '1';
                        error += grades;                                            // The error is as much as toggling from white to black
                    }
                }
                // Now diffuse error:
                if (x < w - 1) bmp[i + 1] += (7 * error) >> 4;
                if (y < h - 1) {
                    bmp[i + w] += (5 * error) >> 4;
                    if (x > 0) bmp[i + w - 1] += (3 * error) >> 4;
                    if (x < w-1) bmp[i + w + 1] += (error) >> 4;
                }
            }
        }
        else {
            for (int x = w - 1; x >= 0; x--) {
                int i = y*w + x;
                if (bmp[i] < threshold) {
                    row[x] = '1';                                               // Output black pixel
                    error = bmp[i];                                             // Positive error: carry-on to neighbors
                } else {
                    row[x] = '0';                                               // Output white pixel
                    error = bmp[i] - grades;                                    // Negative error: to be borrowed from neighbors
                }
                if (tutiHalftone) {
                    if (((x&1) == 1) && (row[x] != '0')) {                      // odd-odd pixel must be white; if it's not fix it
                        row[x] = '0';
                        error -= grades;                                        // The error is as much as toggling from black to white
                    }
                }
                // Now diffuse error:
                if (x > 0) bmp[i - 1] += (7 * error) >> 4;
                if (y < h - 1) {
                    bmp[i + w] += (5 * error) >> 4;
                    if (x > 0) bmp[i + w - 1] += (error) >> 4;
                    if (x < w - 1) bmp[i + w + 1] += (3 * error) >> 4;
                }
            }
        }
        for (int x = 0; x < w; x++) fprintf(pbmfile, "%c\n", row[x]);           // Flush row to output.
    }
    delete[] bmp;
    fclose(pbmfile);
    return 0;
}
